/*
 * lxcall-a.9.c
 * Copyright (C) 2011-2014, 2016, 2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <ct.h>
#include <e4.h>
#include <fd.h>
#include <it.h>
#include <lxcall-defs.h>
#include <lxcall-inter.h>
#include <lxcast.h>
#include <track.h>
#include <trans.h>

extern const void *const x1f4_a1_walk_link;

static int deck_call(void **, const struct x1f4_function_type *, unsigned,
		     struct x1f4_linetext_type *, int);
static int miss_call(void *, int, void *, struct x1f4_dispatch_type **);
static int pick_call(void *, void **, void **,
		     const struct x1f4_function_type *, unsigned, unsigned,
		     struct x1f4_linetext_type *, unsigned, int *);
static int slip_call(void *, void **, void **,
		     const struct x1f4_function_type *, unsigned, unsigned,
		     struct x1f4_linetext_type *, unsigned, int *);

static int
deck_call(void **input, const struct x1f4_function_type *function_data,
	  unsigned offset, struct x1f4_linetext_type *linetext_data, int lock)
{
    int *args;
    unsigned count, flags;
    void **used;

    count = function_data->count - offset;

    flags = linetext_data->function.flags;

    do {
	unsigned class;

	if (flags & (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE)) {
	    flags |= X1f4_E4_POST_TYPE;
	}

	args = (int *) function_data->args;

	class = linetext_data->function.count;
	if (lock < class) {
	    int *line, slip;
	    const int *plot;

	    line = args;
	    plot = linetext_data->function.args;

	    line += lock;
	    plot += lock;

	    slip = *plot;
	    *line = slip == X1f4_E4_SLIP ? function_data->args[lock] : slip;

	    if (flags & X1f4_E4_POST_TYPE) {
		if (linetext_data->function.flags & X1f4_E4_POST_TYPE) {
		    line[count] = plot[class];
		} else {
		    line[count] = 0;
		}
	    }
	}
    } while (0);

    linetext_data->function.args = args;
    linetext_data->function.count = count;
    linetext_data->function.flags =
	flags & ~(X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST);

    used = input[-1];

    used[X1f4_E4_RENDITION] = &linetext_data->function;

    return 0;
}


static int
miss_call(void *text, int success, void *output,
	  struct x1f4_dispatch_type **dispatch)
{
    *dispatch = NULL;

    return success;
}


static int
pick_call(void *output, void **input, void **track,
	  const struct x1f4_function_type *function_data, unsigned offset,
	  unsigned bits, struct x1f4_linetext_type *linetext_data,
	  unsigned norm, int *lock)
{
    int status;

    do {
	if (linetext_data->function.flags
	    & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)) {
	    for (; norm; norm--) {
		status = deck_call
		    (input, function_data, offset, linetext_data, *lock);
		if (status) {
		    break;
		} else {
		    lock++;
		}
	    }
	    if (status) {
		break;
	    }
	}

	status = linetext_data->function.function
	    ((void *) linetext_data->context, output, track);
    } while (0);

    return status;
}


static int
slip_call(void *output, void **input, void **track,
	  const struct x1f4_function_type *function_data, unsigned offset,
	  unsigned bits, struct x1f4_linetext_type *linetext_data,
	  unsigned norm, int *lock)
{
    int status = 0;

    do {
	unsigned turn;

	turn = linetext_data->function.flags;

	{
	    struct x1f4_dispatch_type *dispatch_data;
	    void *subtext;

	    output = input[0];

	    subtext = (void *) linetext_data->context;

	    if (turn & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)) {
		for (; norm; norm--) {
		    status = deck_call
			(track, function_data, offset, linetext_data, *lock);
		    if (status) {
			break;
		    } else {
			lock++;
		    }
		}
		if (status) {
		    break;
		}
	    }

	    dispatch_data = input[2];

	    dispatch_data->back = NULL;
	    dispatch_data->call = miss_call;

	    dispatch_data->transfer.context = subtext;
	    dispatch_data->transfer.function_data = &linetext_data->function;
	    dispatch_data->transfer.input = track;
	    dispatch_data->transfer.output = output;
	}
    } while (0);

    return status;
}


int
_libx1f4i0_lxcall_rail_call(struct x1f4_linetext_type *linetext_data,
			    const struct x1f4_function_type *function_data,
			    void *output, unsigned offset, void **input,
			    unsigned bits, struct x1f4_iturn_type *iturn_data,
			    struct x1f4_cturn_type *cturn_data,
			    struct x1f4_track_type *track_data,
			    const struct x1f4_eelookup_type *eelookup_data,
			    unsigned norm, int *lock,
			    struct fxdata_type *fxdata_data,
			    struct x1f4_dxcast_type *dxcast_data)
{
    int linked, status;
    unsigned i;
    void *meta, **track;

    linked = output == x1f4_a1_walk_link;

    if (linked) {
	track = input[1];
    } else {
	track = input;
    }

    meta = *(track - 1);

    track += offset;

    i = norm;
    while (1) {
	status = _libx1f4i0_lxcall_even_mark
	    (&linetext_data->function, function_data, offset, &track, bits,
	     iturn_data, cturn_data, track_data, eelookup_data, *lock,
	     fxdata_data, dxcast_data);
	if (status) {
	    break;
	} else {
	    i--;
	    if (i) {
		lock++;
		fxdata_data++;
		dxcast_data++;
	    } else {
		break;
	    }
	}
    }
    if (status) {
    } else {
	track[-1] = meta;

	lock -= norm - 1;

	if (linked) {
	    status = slip_call
		(output, input, track, function_data, offset, bits,
		 linetext_data, norm, lock);
	} else {
	    status = pick_call
		(output, input, track, function_data, offset, bits,
		 linetext_data, norm, lock);
	}
    }

    return status;
}
