/*
 * lxroll-a.e.c
 * Copyright (C) 2008-2014, 2016, 2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <errno.h>
#include <string.h>

#include <er.h>
#include <exerrors.h>
#include <lxroll-defs.h>
#include <lxroll-types.h>

#define lxroll(lxroll) \
    ((struct lxroll_type *) (lxroll))

int
_libx1f4i0_lxroll_stat_back(void *lxroll)
{
    LINE_EEER(lxroll(lxroll)->link_e);

    PUSH_DATA("the two connection files to be opened are one and the same",
	      58);

    POST_EEER();

    return X1f4_EX_CAN_CONTINUE;
}


int
_libx1f4i0_lxroll_stat_call(void *lxroll, const char *call)
{
    int error;

    error = errno;

    LINE_EEER(lxroll(lxroll)->link_e);

    PUSH_TEXT(call);
    PUSH_DATA(" failed: ", 9);
    PUSH_TEXT(strerror(error));

    POST_EEER();

    return X1f4_EX_CAN_CONTINUE;
}


int
_libx1f4i0_lxroll_stat_flat(void *lxroll)
{
    LINE_EEER(lxroll(lxroll)->link_e);

    PUSH_DATA("file not opened", 15);

    POST_EEER();

    return X1f4_EX_CAN_CONTINUE;
}


int
_libx1f4i0_lxroll_stat_getf(void *lxroll)
{
    int error;

    error = errno;

    LINE_EEER(lxroll(lxroll)->link_e);

    PUSH_DATA("cannot retrieve the file flags: ", 32);
    PUSH_TEXT(strerror(error));

    POST_EEER();

    return X1f4_EX_CAN_CONTINUE;
}


int
_libx1f4i0_lxroll_stat_info(void *lxroll, const char *node, const char *error)
{
    LINE_EEER(lxroll(lxroll)->link_e);

    PUSH_DATA("cannot get host `", 17);
    PUSH_TEXT(node);
    PUSH_DATA("' address: ", 11);
    PUSH_TEXT(error);

    POST_EEER();

    return X1f4_EX_CAN_CONTINUE;
}


int
_libx1f4i0_lxroll_stat_lock(void *lxroll, const char *node)
{
    LINE_EEER(lxroll(lxroll)->link_e);

    PUSH_DATA("cannot match `", 14);
    PUSH_TEXT(node);
    PUSH_DATA("' address type", 14);

    POST_EEER();

    return X1f4_EX_CAN_CONTINUE;
}


int
_libx1f4i0_lxroll_stat_part(void *lxroll, int part)
{
    int error;

    error = errno;

    LINE_EEER(lxroll(lxroll)->link_e);

    PUSH_DATA("cannot set the ", 15);
    switch (part) {
    case REUSE_ADDRESS:
	PUSH_DATA("reuse address", 13);
	break;
    case TCP_NO_DELAY:
	PUSH_DATA("nodelay", 7);
    }

    PUSH_DATA(" flag: ", 7);
    PUSH_TEXT(strerror(error));

    POST_EEER();

    return X1f4_EX_CAN_CONTINUE;
}


int
_libx1f4i0_lxroll_stat_setf(void *lxroll)
{
    int error;

    error = errno;

    LINE_EEER(lxroll(lxroll)->link_e);

    PUSH_DATA("cannot set the file flags: ", 27);
    PUSH_TEXT(strerror(error));

    POST_EEER();

    return X1f4_EX_CAN_CONTINUE;
}


int
_libx1f4i0_lxroll_stat_side(void *lxroll)
{
    LINE_EEER(lxroll(lxroll)->link_e);

    PUSH_DATA("not a system file", 17);

    POST_EEER();

    return X1f4_EX_CAN_CONTINUE;
}
