/*
 * qsce.o.c
 * Copyright (C) 2016, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <qscc-config.h>

#include <stddef.h>
#include <string.h>

#include <qscc-defs.h>
#include <qscc-inter.h>
#include <qscc-names.h>
#include <qscc-types.h>

#undef __DELAY_SPLIT__
#define __DELAY_SPLIT__			0

#define false(e)			0
#define true(e)				1

#define qsrate(qsrate)			((struct qsrate_type *) (qsrate))

#define roll_node(___n, ___l)		join_node((___n), NULL, (___l))

static int come_node(void *, unsigned, unsigned, unsigned);
static int ever_node(void *, unsigned, unsigned, unsigned);
static int head_rate(struct qsrate_type *, unsigned, unsigned, unsigned,
		     void *);
static int join_node(void *, void *, unsigned);
static int lead_node(void *, void *, unsigned, unsigned, unsigned *);
static int lock_node(void *, struct fpnode_type *, struct fpnode_type *,
		     unsigned, unsigned, unsigned, void *, int);
static int near_node(void *, void *, unsigned, unsigned, unsigned *);
static int pass_node(void *, void *, unsigned, unsigned, unsigned *);
static int pick_cast(struct qsrate_type *, struct qsrate_type *);
static int push_last(struct fpnode_type *, unsigned, unsigned, unsigned, int,
		     unsigned, unsigned);
static int push_post(struct fpnode_type *, unsigned, unsigned, unsigned, int);
static int tail_rate(struct qsrate_type *, unsigned, unsigned, unsigned,
		     void *);

static int
come_node(void *node, unsigned rate, unsigned lock, unsigned size)
{
    ever_node(node, rate, lock, size);
    roll_node(node, lock);

    return 0;
}


static int
ever_node(void *node, unsigned rate, unsigned lock, unsigned size)
{
    struct fpnode_type *fpnode_data;
    unsigned call, i, line, miss, s, v;

    line = 1 << lock;

    fpnode_data = node;
    if (rate == 1) {
	for (i = line; i; i--) {
	    fpnode_data->call = fpnode_data->node;
	    fpnode_data++;
	}
    } else {
	call = size;

	fpnode_data += line;
	for (i = line; i; i--) {
	    s = i - 1;
	    v = 1;
	    while (s & v) {
		call -= (fpnode_data - v)->call;
		v <<= 1;
	    }
	    fpnode_data--;
	    miss = fpnode_data->call;
	    fpnode_data->call = call;
	    call = miss;
	}
    }

    return 0;
}


static int
head_rate(struct qsrate_type *qsrate_data, unsigned debt, unsigned head,
	  unsigned over, void *roll)
{
    int status;
    unsigned count, rate;

    count = qsrate_data->link_a.fplist.size;

    status = 0;

    rate = qsrate_data->link_a.fplist.rate;

    do {
	struct fpnode_type *fpnode_call = NULL, *fpnode_data;
	unsigned line, lock, mall;
	void *star;

	lock = qsrate_data->link_a.fplist.lock;

	if (qsrate_data->link_a.fplist.node >> lock) {
	    status = _libx1f4l2_fall_qsrate(qsrate_data);
	    if (status) {
		break;
	    } else {
		rate++;
	    }
	}

	qsrate_data->link_a.fplist.size = count + head;

	fpnode_data = qsrate_data->link_a.fplist.fpnode;

	if (1) {
	    unsigned half;

	    half = 1 << lock >> 1;

	    while (half) {
		struct fpnode_type *fpnode_slip;
		unsigned pick;

		fpnode_slip = fpnode_data + half;
		pick = fpnode_slip->call;
		if (pick) {
		    fpnode_slip->call = pick + head;
		} else {
		    fpnode_data = fpnode_slip;
		}

		half >>= 1;
	    }

	    rate--;
	}

	while (rate) {
	    struct fpnode_type *fpnode_trip;

	    if (debt == rate) {
		if (over < 1 << lock >> 1) {
		} else {
		    status = lock_node
			(qsrate_data, fpnode_call, fpnode_data, rate, head,
			 over, roll, 0);

		    break;
		}
	    } else {
	    }

	    if (1) {
		struct fpnode_type *fpnode_land;
		unsigned half;

		fpnode_land = fpnode_data->star;
		half = 1 << lock >> 1;
		for (; half; half >>= 1) {
		    fpnode_land[half].call += head;
		}
	    } else {
	    }

	    if (fpnode_data->node >> lock) {
#if __DELAY_SPLIT__
#else
		int excess;
		unsigned index = 0;

		excess = _libx1f4l2_sand_qsrate
		    (qsrate_data, fpnode_call, &fpnode_data, &index,
		     rate, 0);
		if (excess) {
		    if (status) {
		    } else {
			status = excess;
		    }

		    break;
		}
#endif				/* __DELAY_SPLIT__ */
	    }

	    fpnode_trip = fpnode_call;

	    fpnode_call = fpnode_data;

	    fpnode_data = fpnode_data->star;

	    if (debt == rate) {
		if (1) {
		    struct fpnode_type *fpnode_land;
		    unsigned half;

		    fpnode_land = fpnode_call->star;
		    half = 1 << lock >> 1;
		    for (; half; half >>= 1) {
			fpnode_land[half].call -= head;
		    }
		}

		ever_node(roll, rate, lock, head);
		ever_node(fpnode_data, rate, lock, fpnode_data[1].call);

		if (1 << lock < fpnode_call->node + over) {
		    unsigned miss;

		    pass_node(roll, fpnode_data, over, lock, &miss);
		    roll_node(fpnode_data, lock);
		    roll_node(roll, lock);
		    fpnode_call->node -= (1 << lock >> 1) - over;

		    status = lock_node
			(qsrate_data, fpnode_trip, fpnode_call, rate,
			 head + miss, 1 << lock >> 1, roll, 0);

		    break;
		} else {
#define poll				fpnode_data

		    join_node(roll, poll, lock);
		    fpnode_call->node += over;
		    fpnode_call->star = roll;

		    status = qsrate_data->link_m.free
			(qsrate_data->link_m.data, poll);

#undef poll

		    break;
		}
	    }

	    rate--;
	}
	if (rate) {
	    break;
	}

	line = qsrate_data->link_a.fplist.line;

	mall = qsrate_data->link_a.fplist.mall;

	count = fpnode_data->node;

	if (line < count + head) {
	    int excess;
	    struct fpnode_type *fpnode_line;
	    unsigned index = 0, part, trap;

	    part = line - count;

	    star = fpnode_data->star;
	    memmove((char *) star + part * mall, star, count * mall);
	    memcpy(star, (char *) roll + (head - part) * mall, part * mall);

	    head -= part;

	    fpnode_data->node = line;

#if __DELAY_SPLIT__
#endif				/* __DELAY_SPLIT__ */

	    /*
	     * troubling here is that node linking fpnode_data is not exactly
	     * valid (as head was added, the fpnode_data count had become node
	     * + head).
	     */
	    /*
	     * and...
	     * this sequenced insertion is pretty lame.  maddening.
	     */

	    trap = 1 << lock;

	    if (fpnode_call) {
		fpnode_line = fpnode_call->star;
	    } else {
		fpnode_line = qsrate_data->link_a.fplist.fpnode;
	    }

	    head--;

	    part = fpnode_data - fpnode_line;
	    part++;
	    if (part < trap) {
		unsigned land = 1;

		fpnode_line[part].call -= head;
		while (land < trap) {
		    if (part & land) {
			part += land;
			if (part < trap) {
			    fpnode_line[part].call -= head;
			} else {
			    break;
			}
		    }

		    land <<= 1;
		}
	    }

	    excess = _libx1f4l2_mold_qsrate
		(qsrate_data, fpnode_call, &fpnode_data, &index);

	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }

	    part = fpnode_data - fpnode_line;
	    part++;
	    if (part < trap) {
		unsigned land = 1;

		fpnode_line[part].call += head;
		while (land < trap) {
		    if (part & land) {
			part += land;
			if (part < trap) {
			    fpnode_line[part].call += head;
			} else {
			    break;
			}
		    }

		    land <<= 1;
		}
	    }

	    head++;
	}

	count = fpnode_data->node;

	fpnode_data->node = count + head;

	star = fpnode_data->star;
	memmove((char *) star + head * mall, star, count * mall);
	memcpy(star, roll, head * mall);
    } while (0);

    return status;
}


static int
join_node(void *node, void *roll, unsigned lock)
{
    struct fpnode_type *fpnode_data, *fpnode_slip, *fpnode_text;
    unsigned i, line, miss;

    line = 1 << lock;

    fpnode_data = node;
    fpnode_text = node;
    for (i = line; i; i--) {
	if (fpnode_data->node) {
	    *fpnode_text++ = *fpnode_data;
	}
	fpnode_data++;
    }

    if (roll) {
	fpnode_slip = roll;
	for (i = line; i; i--) {
	    if (fpnode_slip->node) {
		*fpnode_text++ = *fpnode_slip;
	    }
	    fpnode_slip++;
	}
    }

    i = fpnode_data - fpnode_text;
    if (line >> 1 < i) {
	for (; i; i--) {
	    fpnode_data--;
	    fpnode_data->call = 0;
	    fpnode_data->node = 0;
	}
    } else {
	for (i = fpnode_data - fpnode_text; i; i--) {
	    fpnode_data--;
	    fpnode_data->call = 0;
	    fpnode_data->node = 0;
	    fpnode_data--;
	    fpnode_text--;
	    *fpnode_data = *fpnode_text;
	}
    }

    fpnode_data = node;
    fpnode_data += line;
    for (i = line - 1; i; i--) {
	fpnode_data--;
	fpnode_data->call = (fpnode_data - 1)->call;
    }
    {
	fpnode_data--;
	fpnode_data->call = 0;
    }

    for (--lock, miss = 2; lock; lock--, miss <<= 1) {
	line = 1 << lock;
	fpnode_data = node;
	fpnode_text = fpnode_data + (miss >> 1);
	for (i = line - 1; i; i--) {
	    fpnode_data += miss;
	    fpnode_data->call += fpnode_text->call;
	    fpnode_text += miss;
	}
    }

    return 0;
}


static int
lead_node(void *node, void *roll, unsigned over, unsigned lock, unsigned *trip)
{
    struct fpnode_type *fpnode_data, *fpnode_text;
    unsigned i, line, miss = 0;

    line = 1 << lock;

    fpnode_data = node;
    fpnode_text = fpnode_data;
    for (i = line; i; i--) {
	if (fpnode_data->node) {
	    *fpnode_text = *fpnode_data;
	    fpnode_text++;
	}
	fpnode_data++;
    }

    fpnode_data = roll;
    i = line - over;
    for (; i;) {
	if (fpnode_data->node) {
	    *fpnode_text = *fpnode_data;
	    fpnode_text++;
	    miss += fpnode_data->call;
	    fpnode_data->node = 0;
	    i--;
	}
	fpnode_data++;
    }

    *trip = miss;

    return 0;
}


static int
lock_node(void *qsrate, struct fpnode_type *fpnode_call,
	  struct fpnode_type *fpnode_data, unsigned rate, unsigned fine,
	  unsigned over, void *side, int affix)
{
    struct fpnode_type *fpnode_line;
    unsigned cell, last = ~0, lock, post = ~0;

    lock = qsrate(qsrate)->link_a.fplist.lock;

    if (fpnode_call) {
	fpnode_call->node++;
	fpnode_line = fpnode_call->star;
    } else {
	qsrate(qsrate)->link_a.fplist.node++;
	fpnode_line = qsrate(qsrate)->link_a.fplist.fpnode;
    }

    cell = fpnode_data - fpnode_line;

    do {
	if (fpnode_call) {
	    struct fpnode_type *fpnode_last, *fpnode_post;
	    unsigned line, mind, slip, text;

	    line = 1 << qsrate(qsrate)->link_a.fplist.lock;

	    fpnode_last = fpnode_line + cell;
	    fpnode_post = fpnode_last;

	    mind = cell >> 1;
	    slip = (line - cell) >> 1;

	    if (cell & 1) {
		fpnode_last -= 2;
		fpnode_post += 2;
	    } else {
		fpnode_last -= 1;
		fpnode_post += 1;
	    }

	    text = mind < slip ? mind : slip;
	    while (text) {
		if (fpnode_post->node) {
		    fpnode_post += 2;
		} else {
		    post = fpnode_post - fpnode_line;
		    if (1) {
			break;
		    }
		}
		if (fpnode_last->node) {
		    fpnode_last -= 2;
		} else {
		    last = fpnode_last - fpnode_line;
		    if (1) {
			break;
		    }
		}

		text--;
	    }

	    if (text) {
	    } else {
		if (mind < slip) {
		    slip -= mind;
		    for (; slip; slip--) {
			if (fpnode_post->node) {
			    fpnode_post += 2;
			} else {
			    post = fpnode_post - fpnode_line;
			    if (1) {
				break;
			    }
			}
		    }
		} else {
		    mind -= slip;
		    for (; mind; mind--) {
			if (fpnode_last->node) {
			    fpnode_last -= 2;
			} else {
			    last = fpnode_last - fpnode_line;
			    if (1) {
				break;
			    }
			}
		    }
		}
	    }
	} else {
	    struct fpnode_type *fpnode_last, *fpnode_post;
	    unsigned line, mind, slip, text;

#if 0
	    if (cell & 3) {
	    } else {
		if (_libx1f4l2_sail_qsrate
		    (qsrate, &fpnode_line, cell, side, index, fpnode)) {
		    break;
		}
	    }
#endif				/* 0 */

	    line = 1 << qsrate(qsrate)->link_a.fplist.lock;

	    fpnode_last = fpnode_line + cell;
	    fpnode_post = fpnode_last;

	    mind = cell;
	    slip = line - cell - 1;

	    if (0) {
	    } else {
		fpnode_last -= 1;
		fpnode_post += 1;
	    }

	    text = mind < slip ? mind : slip;
	    while (text) {
		if (fpnode_post->node) {
		    fpnode_post++;
		} else {
		    post = fpnode_post - fpnode_line;
		    if (1) {
			break;
		    }
		}
		if (fpnode_last->node) {
		    fpnode_last--;
		} else {
		    last = fpnode_last - fpnode_line;
		    if (1) {
			break;
		    }
		}

		text--;
	    }

	    if (text) {
	    } else {
		if (mind < slip) {
		    slip -= mind;
		    for (; slip; slip--) {
			if (fpnode_post->node) {
			    fpnode_post++;
			} else {
			    post = fpnode_post - fpnode_line;
			    if (1) {
				break;
			    }
			}
		    }
		} else {
		    mind -= slip;
		    for (; mind; mind--) {
			if (fpnode_last->node) {
			    fpnode_last--;
			} else {
			    last = fpnode_last - fpnode_line;
			    if (1) {
				break;
			    }
			}
		    }
		}
	    }
	}

	if (~last) {
	    push_last(fpnode_line, cell, last, fine, affix, lock, rate);

	    if (affix) {
	    } else {
		fpnode_data--;
		fpnode_data[1].node = fpnode_data->node;
		fpnode_data[1].star = fpnode_data->star;
	    }
	} else {
	    push_post(fpnode_line, cell, post, fine, affix);

	    if (affix) {
		fpnode_data++;
	    } else {
		fpnode_data[1].node = fpnode_data->node;
		fpnode_data[1].star = fpnode_data->star;
	    }
	}

	fpnode_data->node = over;
	fpnode_data->star = side;
    } while (0);

    return 0;
}


static int
near_node(void *node, void *roll, unsigned over, unsigned lock, unsigned *trip)
{
    struct fpnode_type *fpnode_data, *fpnode_text;
    unsigned i, line, miss = 0;

    line = 1 << lock;

    /*
     * can add only half, and set i to half.  then adjust final i-for
     * accordingly.  ditto for pass_node.
     */
    fpnode_data = node;
    fpnode_data += line;
    fpnode_text = fpnode_data;
    for (i = line; i; i--) {
	fpnode_data--;
	if (fpnode_data->node) {
	    fpnode_text--;
	    *fpnode_text = *fpnode_data;
	}
    }

    fpnode_data = roll;
    fpnode_data += line;
    i = (line >> 1) - over;
    for (; i;) {
	fpnode_data--;
	if (fpnode_data->node) {
	    fpnode_text--;
	    *fpnode_text = *fpnode_data;
	    miss += fpnode_data->call;
	    fpnode_data->node = 0;
	    i--;
	}
    }

    for (i = line >> 1; i; i--) {
	fpnode_text--;
	fpnode_text->node = 0;
    }

    *trip = miss;

    return 0;
}


static int
pass_node(void *node, void *roll, unsigned over, unsigned lock, unsigned *trip)
{
    struct fpnode_type *fpnode_data, *fpnode_text;
    unsigned i, line, miss = 0;

    line = 1 << lock;

    fpnode_data = node;
    fpnode_text = fpnode_data;
    for (i = line; i; i--) {
	if (fpnode_data->node) {
	    *fpnode_text++ = *fpnode_data;
	}
	fpnode_data++;
    }

    fpnode_data = roll;
    i = (line >> 1) - over;
    for (; i;) {
	if (fpnode_data->node) {
	    *fpnode_text++ = *fpnode_data;
	    miss += fpnode_data->call;
	    fpnode_data->node = 0;
	    i--;
	}
	fpnode_data++;
    }

    for (i = line >> 1; i; i--) {
	fpnode_text->node = 0;
	fpnode_text++;
    }

    *trip = miss;

    return 0;
}


static int
pick_cast(struct qsrate_type *qsrate_data, struct qsrate_type *qsrate_text)
{
    int status;
    unsigned file, line, mall, mile, pipe;

    pipe = qsrate_data->link_a.fplist.size;
    file = qsrate_text->link_a.fplist.size;

    mall = qsrate_data->link_a.fplist.mall;

    mile = file + pipe;
    line = qsrate_data->link_a.fplist.line;
    if (line * 3 >> 1 < mile) {
	void *star;

#if __MICRO_RATES__
	while (~line & 2) {
	    line >>= 1;
	}
#endif				/* __MICRO_RATES__ */
	line *= 3;
	line >>= 1;

	status = qsrate_data->link_m.link
	    (qsrate_data->link_m.data, &star, mall * line);
	if (status) {
	    status = LINK_ERROR;
	} else {
	    int excess;
	    void *data, *roll;

	    data = qsrate_data->link_a.fplist.fpnode;
	    roll = data;

	    qsrate_data->link_a.fplist.fpnode = star;
	    qsrate_data->link_a.fplist.size = 0;

	    status = _libx1f4l2_star_qsrate(qsrate_data, mile);
	    if (status) {
		_libx1f4l2_high_qsrate(qsrate_data);

		qsrate_data->link_a.fplist.fpnode = data;
		qsrate_data->link_a.fplist.size = pipe;
	    } else {
		struct fpnode_type *fpnode_data;

		qsrate_text->link_a.fplist.size = 0;

		fpnode_data = qsrate_data->link_a.fplist.fpnode;
		while (mile) {
		    unsigned node;

		    node = fpnode_data->node;
		    if (node) {
			unsigned call;

			star = fpnode_data->star;

			if (file < mile) {
			    unsigned over;

			    call = pipe < node ? pipe : node;
			    memcpy(star, roll, call * mall);
			    pipe -= call;
			    if (pipe) {
				roll = (char *) roll + call * mall;
			    } else {
				roll = qsrate_text->link_a.fplist.fpnode;
			    }
			    over = node - call;
			    if (over) {
				memcpy((char *) star + call * mall, roll,
				       over * mall);
				roll = (char *) roll + over * mall;
			    }
			} else {
			    call = mile < node ? mile : node;
			    memcpy(star, roll, call * mall);
			    if (1) {
				roll = (char *) roll + call * mall;
			    }
			}

			mile -= node;
		    }

		    fpnode_data++;
		}
	    }

	    excess = qsrate_data->link_m.free(qsrate_data->link_m.data, data);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}
    } else {
	void *data, *text;

	data = qsrate_data->link_a.fplist.fpnode;
	text = qsrate_text->link_a.fplist.fpnode;

#if __MICRO_RATES__
	do {
	    unsigned park, toll;

	    toll = line * 3;

	    while (~line & 2) {
		line >>= 1;
	    }
	    line *= 3;
	    line >>= 1;

	    park = toll;
	    if (pipe < line) {
		toll = line;
	    } else {
		while (!(toll < pipe) && ~toll & 1) {
		    toll >>= 1;
		}
		toll <<= 1;
	    }
	    if (mile < line) {
		park = line;
	    } else {
		while (!(park < mile) && ~park & 1) {
		    park >>= 1;
		}
		park <<= 1;
	    }

	    if (toll < park) {
		status = qsrate_data->link_m.mode
		    (qsrate_data->link_m.data, &data, mall * park);
		if (status) {
		    status = MODE_ERROR;
		    if (1) {
			break;
		    }
		} else {
		    qsrate_data->link_a.fplist.fpnode = data;
		}
	    }
#endif				/* __MICRO_RATES__ */

	    memcpy((char *) data + pipe * mall, text, file * mall);

	    qsrate_data->link_a.fplist.size = mile;
	    qsrate_text->link_a.fplist.size = 0;
#if __MICRO_RATES__
	} while (0);
#endif				/* __MICRO_RATES__ */
    }

#if __MICRO_RATES__
    if (status) {
    } else {
	unsigned park;

	park = qsrate_data->link_a.fplist.line * 3;
	if (file < line) {
	    park = line;
	} else {
	    while (!(park < file) && ~park & 1) {
		park >>= 1;
	    }
	    park <<= 1;
	}

	if (line < park) {
	    status = qsrate_data->link_m.mode
		(qsrate_data->link_m.data, &qsrate_text->link_a.fplist.fpnode,
		 mall * line);
	    if (status) {
		status = MODE_ERROR;
	    }
	}
    }
#endif				/* __MICRO_RATES__ */

    return status;
}


static int
push_last(struct fpnode_type *fpnode_data, unsigned cell, unsigned last,
	  unsigned fine, int affix, unsigned lock, unsigned rate)
{
    unsigned e, l, node, q, w, x;
    void *star;

    fpnode_data += cell;

    x = cell;

    if (affix) {
	if (x & 1) {
	    if ((x + 1) ^ 1 << lock) {
		fpnode_data++;

		w = fpnode_data->call;

		if (1) {
		    e = ~(x + 1);

		    l = 1;
		    while (e & 1) {
			w -= (fpnode_data - l)->call;
			l <<= 1;
			e >>= 1;
		    }
		}

		fpnode_data--;
	    } else {
		_libx1f4l2_late_qsrate(fpnode_data, rate, lock, &w);
	    }
	} else {
	    w = (fpnode_data + 1)->call;
	}

	if ((cell + 1) ^ 1 << lock) {
	    w -= fine;
	}
    } else {
	w = fine;
    }

    node = fpnode_data->node;
    star = fpnode_data->star;

    if (1) {
	q = fpnode_data->call;
    } else {
    }

    {
	if (x & 1) {
	} else {
	    e = x;
	    e--;
	    e &= x;
	    e ^= x;
	    e >>= 1;

	    do {
		(fpnode_data + e)->call -= w;
		e >>= 1;
	    } while (e);
	}

	{
	    fpnode_data->call += w;
	}
    }

    last++;
    last = x - last;
    if (last) {
	unsigned b, i;

	if (x & 1) {
	    unsigned Node;
	    void *Star;

	    fpnode_data--;

	    x--;

	    w = q;

	    q = fpnode_data->call;

	    e = x;
	    e--;
	    e &= x;
	    e ^= x;
	    e >>= 1;

	    b = e;

	    do {
		(fpnode_data + e)->call -= w;
		e >>= 1;
	    } while (e);

	    fpnode_data->call += w;

	    Node = fpnode_data->node;
	    Star = fpnode_data->star;
	    fpnode_data->node = node;
	    fpnode_data->star = star;
	    node = Node;
	    star = Star;

	    last--;
	} else {
	    e = x;
	    e--;
	    e &= x;
	    e ^= x;
	    e >>= 1;

	    b = e;
	}

	i = last >> 1;
	for (; i; i--) {
	    unsigned Node;
	    void *Star;

	    fpnode_data--;

	    x--;

	    w = q;

	    q = fpnode_data->call;

	    fpnode_data++;

	    e = b;

	    do {
		w -= (fpnode_data - e)->call;
		e >>= 1;
	    } while (e);

	    fpnode_data--;

	    fpnode_data->call += w;

	    Node = fpnode_data->node;
	    Star = fpnode_data->star;
	    fpnode_data->node = node;
	    fpnode_data->star = star;

	    fpnode_data--;

	    x--;

	    w = q;

	    q = fpnode_data->call;

	    e = x;
	    e--;
	    e &= x;
	    e ^= x;
	    e >>= 1;

	    b = e;

	    do {
		(fpnode_data + e)->call -= w;
		e >>= 1;
	    } while (e);

	    fpnode_data->call += w;

	    node = fpnode_data->node;
	    star = fpnode_data->star;
	    fpnode_data->node = Node;
	    fpnode_data->star = Star;
	}

	if (last & 1) {
	    unsigned Node;
	    void *Star;

	    fpnode_data--;

	    x--;

	    w = q;

	    q = fpnode_data->call;

	    fpnode_data++;

	    e = b;

	    do {
		w -= (fpnode_data - e)->call;
		e >>= 1;
	    } while (e);

	    fpnode_data--;

	    fpnode_data->call += w;

	    Node = fpnode_data->node;
	    Star = fpnode_data->star;
	    fpnode_data->node = node;
	    fpnode_data->star = star;
	    node = Node;
	    star = Star;
	}
    }

    {
	fpnode_data--;

	{
	    fpnode_data->node = node;
	    fpnode_data->star = star;
	}
    }

    return 0;
}


static int
push_post(struct fpnode_type *fpnode_data, unsigned cell, unsigned post,
	  unsigned fine, int affix)
{
    unsigned e, l, node, w, x;
    void *star;

    fpnode_data += cell;

    x = cell;

    if (x & 1) {
	fpnode_data++;

	w = fpnode_data->call;

	if (1) {
	    e = ~(x + 1);

	    l = 1;
	    while (e & 1) {
		w -= (fpnode_data - l)->call;
		l <<= 1;
		e >>= 1;
	    }
	}

	fpnode_data--;
    } else {
	w = (fpnode_data + 1)->call;
    }

    if (affix) {
	e = w - fine;
	w = fine;
    } else {
	e = fine;
	w -= e;
    }

    node = fpnode_data->node;
    star = fpnode_data->star;

    {
	fpnode_data++;

	if (x & 1) {
	    e = ~x;
	    e = e ^ (e & (e - 1));
	    e >>= 1;

	    do {
		(fpnode_data + e)->call += w;
		e >>= 1;
	    } while (e);
	}
    }

    post--;
    post -= x;
    if (post) {
	unsigned i;

	if (x & 1) {
	    unsigned Node, v;
	    void *Star;

	    x++;

	    Node = fpnode_data->node;
	    Star = fpnode_data->star;
	    fpnode_data->node = node;
	    fpnode_data->star = star;
	    node = Node;
	    star = Star;

	    fpnode_data->call -= w;

	    fpnode_data++;

	    v = fpnode_data->call;

	    v -= w;

	    w = v;

	    post--;
	}

	i = post >> 1;
	for (; i; i--) {
	    unsigned Node, b, v;
	    void *Star;

	    x++;

	    Node = fpnode_data->node;
	    Star = fpnode_data->star;
	    fpnode_data->node = node;
	    fpnode_data->star = star;

	    fpnode_data->call -= w;

	    fpnode_data++;

	    v = fpnode_data->call;

	    e = x;

	    x++;

	    e &= x;
	    e ^= x;
	    e >>= 1;

	    b = e;

	    do {
		v -= (fpnode_data - e)->call;
		e >>= 1;
	    } while (e);

	    v -= w;

	    e = b;

	    do {
		(fpnode_data + e)->call += v;
		e >>= 1;
	    } while (e);

	    node = fpnode_data->node;
	    star = fpnode_data->star;
	    fpnode_data->node = Node;
	    fpnode_data->star = Star;

	    fpnode_data->call -= v;

	    fpnode_data++;

	    w = fpnode_data->call;

	    w -= v;
	}

	/*
	 * not getting into the next too often
	 */
	if (post & 1) {
	    unsigned Node, b, v;
	    void *Star;

	    x++;

	    Node = fpnode_data->node;
	    Star = fpnode_data->star;
	    fpnode_data->node = node;
	    fpnode_data->star = star;
	    node = Node;
	    star = Star;

	    fpnode_data->call -= w;

	    fpnode_data++;

	    v = fpnode_data->call;

	    e = x + 1;
	    e = e ^ (e & x);
	    e >>= 1;

	    b = e;

	    do {
		v -= (fpnode_data - e)->call;
		e >>= 1;
	    } while (e);

	    v -= w;

	    e = b;

	    do {
		(fpnode_data + e)->call += v;
		e >>= 1;
	    } while (e);

	    w = v;
	}
    }

    {
	fpnode_data->call -= w;

	if (affix) {
	} else {
	    fpnode_data->node = node;
	    fpnode_data->star = star;
	}
    }

    return 0;
}


static int
tail_rate(struct qsrate_type *qsrate_data, unsigned debt, unsigned tail,
	  unsigned over, void *roll)
{
    int status;
    unsigned count, rate;

    count = qsrate_data->link_a.fplist.size;

    status = 0;

    rate = qsrate_data->link_a.fplist.rate;

    do {
	struct fpnode_type *fpnode_call = NULL, *fpnode_data;
	unsigned line, lock, mall;

	lock = qsrate_data->link_a.fplist.lock;

	if (qsrate_data->link_a.fplist.node >> lock) {
	    status = _libx1f4l2_fall_qsrate(qsrate_data);
	    if (status) {
		break;
	    } else {
		rate++;
	    }
	}

	qsrate_data->link_a.fplist.size = count + tail;

	fpnode_data = qsrate_data->link_a.fplist.fpnode;

	if (1) {
	    unsigned half, index;

	    half = 1 << lock >> 1;

	    index = count;

	    while (half) {
		struct fpnode_type *fpnode_slip;
		unsigned pick;

		fpnode_slip = fpnode_data + half;
		pick = fpnode_slip->call;
		if (pick ^ index) {
		    index -= pick;
		    fpnode_data = fpnode_slip;
		} else {
		    fpnode_slip->call = pick + tail;
		}

		half >>= 1;
	    }

	    rate--;
	}

	while (rate) {
	    int recall = 0;
	    struct fpnode_type *fpnode_slip, *fpnode_trip;

	    if (debt == rate) {
		if (over < 1 << lock >> 1) {
		} else {
		    status = lock_node
			(qsrate_data, fpnode_call, fpnode_data, rate, tail,
			 over, roll, 1);

		    break;
		}
	    } else {
	    }

	    if (fpnode_data->node >> lock) {
#if __DELAY_SPLIT__
#else
		int excess;
		unsigned index = ~0;

		excess = _libx1f4l2_sand_qsrate
		    (qsrate_data, fpnode_call, &fpnode_data, &index,
		     rate, tail);
		if (excess) {
		    if (status) {
		    } else {
			status = excess;
		    }

		    break;
		}
#endif				/* __DELAY_SPLIT__ */
	    }

	    fpnode_trip = fpnode_call;

	    fpnode_call = fpnode_data;

	    fpnode_data = fpnode_data->star;

	    fpnode_slip = fpnode_data;

	    fpnode_data += 1 << lock;
	    fpnode_data--;
	    if (fpnode_data->node) {
	    } else {
		recall = 1;
		fpnode_data->call += tail;
		fpnode_data--;
	    }

	    if (debt == rate) {
		unsigned rail;

		if (recall) {
		    fpnode_data[1].call -= tail;
		}

		ever_node(roll, rate, lock, tail);
		if (rate ^ 1) {
		    _libx1f4l2_late_qsrate(fpnode_call, rate, lock, &rail);
		} else {
		}
		ever_node(fpnode_slip, rate, lock, rail);

		if (1 << lock < fpnode_call->node + over) {
		    unsigned miss;

		    near_node(roll, fpnode_slip, over, lock, &miss);
		    roll_node(fpnode_slip, lock);
		    roll_node(roll, lock);
		    fpnode_call->node -= (1 << lock >> 1) - over;
		    status = lock_node
			(qsrate_data, fpnode_trip, fpnode_call, rate,
			 tail + miss, 1 << lock >> 1, roll, 1);

		    break;
		} else {
		    join_node(fpnode_slip, roll, lock);
		    fpnode_call->node += over;

		    status = qsrate_data->link_m.free
			(qsrate_data->link_m.data, roll);

		    break;
		}
	    }

	    rate--;
	}
	if (rate) {
	    break;
	}

	line = qsrate_data->link_a.fplist.line;

	mall = qsrate_data->link_a.fplist.mall;

	count = fpnode_data->node;

	if (line < count + tail) {
	    int excess;
	    struct fpnode_type *fpnode_line;
	    unsigned index = ~0, part, trap;

	    memcpy((char *) fpnode_data->star + count * mall, roll,
		   (line - count) * mall);

	    roll = (char *) roll + (line - count) * mall;
	    tail -= line - count;

	    fpnode_data->node = line;

#if __DELAY_SPLIT__
#endif				/* __DELAY_SPLIT__ */

	    /*
	     * troubling here is that node linking fpnode_data is not exactly
	     * valid (as tail was added, the fpnode_data count had become node
	     * + tail).
	     */
	    /*
	     * and...
	     * this sequenced insertion is pretty lame.  maddening.
	     */

	    trap = 1 << lock;

	    if (fpnode_call) {
		fpnode_line = fpnode_call->star;
	    } else {
		fpnode_line = qsrate_data->link_a.fplist.fpnode;
	    }

	    tail--;

	    part = fpnode_data - fpnode_line;
	    part++;
	    if (part < trap) {
		unsigned land = 1;

		fpnode_line[part].call -= tail;
		while (land < trap) {
		    if (part & land) {
			part += land;
			if (part < trap) {
			    fpnode_line[part].call -= tail;
			} else {
			    break;
			}
		    }

		    land <<= 1;
		}
	    }

	    excess = _libx1f4l2_mold_qsrate
		(qsrate_data, fpnode_call, &fpnode_data, &index);

	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }

	    part = fpnode_data - fpnode_line;
	    part++;
	    if (part < trap) {
		unsigned land = 1;

		fpnode_line[part].call += tail;
		while (land < trap) {
		    if (part & land) {
			part += land;
			if (part < trap) {
			    fpnode_line[part].call += tail;
			} else {
			    break;
			}
		    }

		    land <<= 1;
		}
	    }

	    tail++;
	}

	count = fpnode_data->node;

	fpnode_data->node = count + tail;

	memcpy((char *) fpnode_data->star + count * mall, roll, tail * mall);
    } while (0);

    return status;
}


int
x1f4_join_qsrate(void *qsdata, void *qstext)
{
    int status;
    unsigned date, gate;

    date = qsrate(qsdata)->link_a.fplist.rate;
    gate = qsrate(qstext)->link_a.fplist.rate;

    if (date < gate) {
	if (true(gate)) {
	    if (date) {
		unsigned ever, lock, spin;
		void *text;

		lock = qsrate(qstext)->link_a.fplist.lock;

		spin = qsrate(qsdata)->link_a.fplist.size;
		ever = qsrate(qsdata)->link_a.fplist.node;

		text = qsrate(qsdata)->link_a.fplist.fpnode;

		if (false(date == gate)) {
		} else {
		    if (ever < 1 << lock >> 1) {
		    } else {
			come_node(text, date, lock, spin);
		    }

		    status = head_rate(qstext, date, spin, ever, text);
		}

		if (status) {
		} else {
		    qsrate(qsdata)->link_a.fplist.rate = 0;
		    qsrate(qsdata)->link_a.fplist.fpnode =
			qsrate(qsdata)->link_e.fplink_text;
		    qsrate(qsdata)->link_e.fplink_text = NULL;

		    status = _libx1f4l2_high_qsrate(qsdata);
		    if (status) {
		    } else {
			qsrate(qsdata)->link_e.fplink_data = NULL;
		    }
		}
	    } else {
		unsigned deck, head, mall;
		void *roll;

		roll = qsrate(qsdata)->link_a.fplist.fpnode;

		deck = qsrate(qsdata)->link_a.fplist.line;
		mall = qsrate(qsdata)->link_a.fplist.mall;
		head = qsrate(qsdata)->link_a.fplist.size;

		/*
		 * transfer (line / 2 - 1) at most once, as _mold_ may leave no
		 * more than that many slots available after splitting.
		 */
		deck >>= 1;
		deck--;

		status = 0;

		while (head) {
		    unsigned call;

		    call = deck < head ? deck : head;
		    head -= call;
		    status = head_rate
			(qstext, 0, call, call, (char *) roll + head * mall);
		    if (status) {
			head += call;
			break;
		    }
		}

		if (status) {
		    qsrate(qsdata)->link_a.fplist.size = head;
		} else {
		    status = _libx1f4l2_high_qsrate(qsdata);
		}
	    }

	    if (status) {
	    } else {
		status = _libx1f4l2_swap_qsrate(qsdata, qstext);
	    }
	} else {
	}
    } else {
	if (date) {
	    if (gate) {
		unsigned ever, lock, spin;
		void *text;

		lock = qsrate(qsdata)->link_a.fplist.lock;

		spin = qsrate(qstext)->link_a.fplist.size;
		ever = qsrate(qstext)->link_a.fplist.node;

		text = qsrate(qstext)->link_a.fplist.fpnode;

		if (date == gate) {
		    unsigned lead, miss, pass;
		    void *data;

		    data = qsrate(qsdata)->link_a.fplist.fpnode;

		    lead = qsrate(qsdata)->link_a.fplist.node;
		    pass = qsrate(qsdata)->link_a.fplist.size;

		    ever_node(data, date, lock, pass);
		    ever_node(text, date, lock, spin);

		    if (1 << lock < lead + ever) {
			lead_node(data, text, lead, lock, &miss);
			roll_node(data, lock);
			roll_node(text, lock);

			ever += lead;
			ever -= 1 << lock;

			spin -= miss;

			qsrate(qsdata)->link_a.fplist.node = 1 << lock;
			qsrate(qsdata)->link_a.fplist.size = pass + miss;

			qsrate(qstext)->link_a.fplist.node = ever;
			qsrate(qstext)->link_a.fplist.size = spin;

			status = _libx1f4l2_fall_qsrate(qsdata);
			if (status) {
			} else {
			    if (ever < 1 << lock >> 1) {
			    } else {
				come_node(text, gate, lock, spin);
			    }

			    status = tail_rate
				(qsdata, gate, spin, ever, text);
			}
		    } else {
			join_node(data, text, lock);

			qsrate(qsdata)->link_a.fplist.node = lead + ever;
			qsrate(qsdata)->link_a.fplist.size = pass + spin;

			status = qsrate(qstext)->link_m.free
			    (qsrate(qstext)->link_m.data, text);
		    }
		} else {
		    if (ever < 1 << lock >> 1) {
		    } else {
			come_node(text, gate, lock, spin);
		    }

		    status = tail_rate(qsdata, gate, spin, ever, text);
		}

		if (status) {
		} else {
		    qsrate(qstext)->link_a.fplist.rate = 0;
		    qsrate(qstext)->link_a.fplist.fpnode =
			qsrate(qstext)->link_e.fplink_text;
		    qsrate(qstext)->link_e.fplink_text = NULL;

		    status = _libx1f4l2_high_qsrate(qstext);
		    if (status) {
		    } else {
			qsrate(qstext)->link_e.fplink_data = NULL;
		    }
		}
	    } else {
		unsigned deck, mall, tail;
		void *roll;

		roll = qsrate(qstext)->link_a.fplist.fpnode;

		deck = qsrate(qstext)->link_a.fplist.line;
		mall = qsrate(qstext)->link_a.fplist.mall;
		tail = qsrate(qstext)->link_a.fplist.size;

		/*
		 * transfer (line / 2 - 1) at most once, as _mold_ may leave no
		 * more than that many slots available after splitting.
		 */
		deck >>= 1;
		deck--;

		status = 0;

		while (tail) {
		    unsigned call;

		    call = deck < tail ? deck : tail;
		    status = tail_rate(qsdata, 0, call, call, roll);
		    if (status) {
			break;
		    } else {
			tail -= call;
			roll = (char *) roll + call * mall;
		    }
		}

		if (status) {
		    qsrate(qsdata)->link_a.fplist.size = tail;
		    memmove
			(qsrate(qstext)->link_a.fplist.fpnode, roll,
			 tail * mall);
		} else {
		    status = _libx1f4l2_high_qsrate(qstext);
		}
	    }
	} else {
	    status = pick_cast(qsdata, qstext);
	}
    }

    return status;
}
